module eta_pspi
  
  use ani_type
  use ani_parms
  use FFT
  
  implicit none
  
contains

! =================================================================================
! Main pspi module
  subroutine etapspi(iwa, inode, sign, ws, model, refr, gat)
    
    character*3					:: gat
    integer                                     :: iwa, inode, sign
    complex                                     :: ws(:)
    complex, allocatable                        :: cube(:,:)
    real                                        :: model(:,:), refr(:,:)

    real                                        :: v, eta
    integer                                     :: nref, iref

    allocate(cube(nx,nref_max))

    do iref=1, nref_max
       cube(:,iref) = ws
    end do

    call select_ref(refr(:,1), nref)
    
    do iref = 1, nref
       v   = refr(iref,1)
       eta = refr(iref,2)
       call eta_ssf_exp(iwa, cube(:,iref), model(:,1), v, eta, sign)
    end do
    
    call interp_lloyd(cube, ws, model, refr, scales, nref)

    deallocate(cube)
  end subroutine etapspi

! =================================================================================
! One step in depth extrapolation
  subroutine eta_ssf_exp(iwa, DATw, vxmodel, v, eta, sign)
    implicit none
    complex  :: DATw(:)
    real     :: vxmodel(:)
    real     :: dw, dkx, w
    real     :: wov, wov2, kx, kx2, kz, kz2, delta_s
    integer  :: ikx, jkx, iwa, ix, nkx, sign
    
    real     :: v, eta
    real     :: phase, time

    nkx = kxm%n
    dkx = kxm%d * 2*pi

    call FFT1D(DATw, nx, 1)
    
    w = rec%w%all(iwa)
    if(w.eq.0.0) w=1.0e-06
    
    wov=w/v
    wov2=wov*wov
    
    do ikx = 1, nkx
       kx = kxm%all(ikx)
       kx2 = kx * kx
       kz2 = wov2*(wov2-kx2*(1.0+2.0*eta))/ &
             (wov2-2.0*kx2*eta)       
       if(kz2.ge.0.0) then
          kz=sign*sqrt(kz2)
          phase=-kz*dz
          DATw(ikx)=DATw(ikx)*cmplx(cos(phase), sin(phase))
       else
          DATw(ikx)=cmplx(0.0,0.0)
       end if       
    end do
    
    call FFT1D(DATw, nx, -1)
    
   do ix=1, nx
      delta_s=1.0/vxmodel(ix)-1.0/v
      time=-w*dz*delta_s*sign
      DATw(ix)=DATw(ix)*cmplx(cos(time), sin(time))
   end do
    
  end subroutine eta_ssf_exp
  
! =================================================================================
! Lloyd interpolation
  subroutine interp_lloyd(cube, wavefield, ani_modx, ani_refx, scale, nref)
                     
    implicit none
    complex :: cube(:,:)
    complex :: wavefield(:)
    real :: ani_modx(:,:), ani_refx(:,:), scale(:)
    real, allocatable :: cent_modx(:), cent_refx(:)

    real    :: cmin, cmax, dc, error, error_min
    integer :: nref, lc, ipar, ix, iref

    if(nref==1)then
       do ix=1, nx
          wavefield(ix)=cube(ix, 1)
       end do
       return
    end if

    allocate(cent_modx(nx), cent_refx(nref))

    cent_modx=0.0
    cent_refx=0.0

    do ipar=1, npar
   
       cent_modx=cent_modx + ani_modx(:,ipar)*scale(ipar)
       cent_refx=cent_refx + ani_refx(:,ipar)*scale(ipar)

    end do
    
    cent_modx=cent_modx/npar
    cent_refx=cent_refx/npar

    do ix=1, nx

       error_min=abs(cent_modx(ix)-cent_refx(1))
       lc=1
       do iref=2, nref
          
          error=abs(cent_modx(ix)-cent_refx(iref))
          if(error<error_min)then
             error_min=error
             lc=iref
          end if

       end do
       wavefield(ix)=cube(ix, lc)
    end do

    deallocate(cent_modx, cent_refx)
    return
  end subroutine interp_lloyd


! =================================================================================
! Reference model selection module
  subroutine select_ref(array, ncount)
    implicit none
    integer :: iref, nref_max, ncount
    real    :: array(:)

    nref_max=size(array)
    
    if(array(1)<=0.0)then
       ncount=1
       return
    else
       ncount=0
       do iref=1, nref_max
          ncount=ncount+1
          if(array(iref)<=0.0)then
             ncount=iref-1
             return
          end if
       end do
    end if
    return
  end subroutine select_ref

end module eta_pspi
